# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-02-05 13:49
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chaos', '0005_auto_20180119_1257'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprise',
            name='geometry',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=4326, verbose_name='Geometry'),
        ),
        migrations.AddField(
            model_name='enterprise',
            name='geometry_900913',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=900913, verbose_name='Geometry 900913'),
        ),
        migrations.AddField(
            model_name='project',
            name='geometry',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=4326, verbose_name='Geometry'),
        ),
        migrations.AddField(
            model_name='project',
            name='geometry_900913',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=900913, verbose_name='Geometry 900913'),
        ),
        migrations.AddField(
            model_name='task',
            name='geometry',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=4326, verbose_name='Geometry'),
        ),
        migrations.AddField(
            model_name='task',
            name='geometry_900913',
            field=django.contrib.gis.db.models.fields.MultiPolygonField(null=True, srid=900913, verbose_name='Geometry 900913'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import chaos.models
import django.db.models.deletion
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('chaos', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, verbose_name='Type', to='chaos.ProjectType'),
        ),
        migrations.AlterField(
            model_name='task',
            name='responsible',
            field=models.ForeignKey(related_name='tasks', on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.ForeignKey(related_name='tasks', on_delete=django.db.models.deletion.PROTECT, default=chaos.models.default_task_status, verbose_name='Status', to='chaos.TaskStatus'),
        ),
    ]



# coding: utf-8
"""
This module is responsible to export
tasks from projects, by user, by project
from the django models.
"""
from __future__ import unicode_literals
from collections import OrderedDict
import os
import codecs
import logging
import pyexcel
from wsgiref.util import FileWrapper
from django.utils.translation import ugettext as _
from django.http import HttpResponse
from django.conf import settings
from rest_framework.reverse import reverse
from icalendar.cal import Event, Calendar
from icalendar import vCalAddress, vText
from .rest import Responses
logger = logging.getLogger(__name__)


class TaskExport(object):

    publish_url = True
    default_filename = None
    request = None
    content_type = 'text/text'

    def __init__(self,
                 publish_url=True,
                 default_filename='task-export.txt',
                 request=None):

        self.publish_url = publish_url
        self.default_filename = default_filename
        self.request = request

    def _get_task_url(self, task):

        if self.publish_url:
            return reverse('task-detail', kwargs={'pk': task.pk}, request=self.request)
        else:
            return ''

    def get_download_name(self, output_path):
        folder, filename = os.path.split(output_path)
        return filename

    def get_output_path(self):
        from .utils import get_output_path as op
        output_path = os.path.join(settings.MEDIA_ROOT, 'export', 'chaos')
        chaos_export_path = settings.CHAOS_EXPORT_PATH if hasattr(settings, 'CHAOS_EXPORT_PATH') else output_path
        return op(chaos_export_path, self.default_filename)

    def export(self, queryset):
        raise NotImplementedError

    def respond(self, queryset):
        raise NotImplementedError


class XLSXTaskExport(TaskExport):

    publish_url = True
    request = None
    content_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    default_filename = 'tasks.xlsx'

    def __init__(self,
                 publish_url=True,
                 default_filename='tasks.xlsx',
                 request=None):
        super(XLSXTaskExport, self).__init__(
            publish_url=publish_url,
            default_filename=default_filename,
            request=request)

    def _serialize_task(self, task):
        parent = task.parent.id if task.parent else None
        created_by = task.created_by.get_username() if task.created_by else None
        responsible = task.responsible.get_username() if task.responsible else None
        description = task.description if task.description else None
        return OrderedDict([
            (_('id'), task.pk),
            (_('UUID'), str(task.uuid)),
            (_('URL'), self._get_task_url(task)),
            (_('Project'), task.project.name),
            (_('Created By'), created_by),
            (_('Date Created'), task.date_created),
            (_('Date Updated'), task.date_updated),
            (_('Name'), task.name),
            (_('Description'), description),
            (_('Parent'), parent),
            (_('Responsible'), responsible),
            (_('Status'), task.status.code),
            (_('Status Description'), task.status.description),
            (_('Weight'), task.weight),
            (_('Start Date'), task.start_date),
            (_('End Date'), task.end_date),
            (_('# Comments'), task.comments.all().count()),
            (_('# Children'), task.children.all().count())
        ])

    def export(self, queryset):

        records = [self._serialize_task(task) for task in queryset]
        output_path = self.get_output_path()
        pyexcel.save_as(
            records=records,
            dest_file_name=output_path,
            dest_sheet_name='tasks'
        )
        return output_path

    def respond(self, queryset):

        """
        Creates a response with
        the export content and returns it
        """
        try:
            logger.info('exporting tasks to xlsx')
            output = self.export(queryset)
            download_name = self.get_download_name(output)
            with open(output, 'rb') as outfile:
                response = HttpResponse(FileWrapper(outfile), content_type=self.content_type)  # noqa
                response['Content-Disposition'] = 'attachment; filename=%s' % download_name
                return response
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while exporting to xlsx. error: %s', message, exc_info=True)
            return Responses.server_error()


class ICALTaskExport(TaskExport):

    """
    Exports tasks to iCal
    """

    publish_url = True
    request = None
    content_type = 'text/calendar'
    default_filename = 'tasks.ics'

    def __init__(self,
                 publish_url=True,
                 default_filename='tasks.ics',
                 request=None):
        super(ICALTaskExport, self).__init__(
            publish_url=publish_url,
            default_filename=default_filename,
            request=request)

    def _create_attendee(self, user, role=None):

        if not role:
            role = 'REQ-PARTICIPANT'

        attendee = vCalAddress('MAILTO:{0}'.format(user.email))
        attendee.params['cn'] = vText(user.get_username())
        attendee.params['role'] = role
        return attendee

    def _create_event(self, task):
        event = Event()
        event.add('uid', task.uuid)
        event.add('summary', task.name)
        event.add('description', task.description)
        event.add('dtstamp', task.date_created)
        if self.publish_url:
            event.add('url', self._get_task_url(task))
        if task.start_date:
            event.add('dtstart', task.start_date)
        if task.end_date:
            event.add('dtend', task.end_date)

        already_included_users = []

        if task.created_by:
            organizer = self._create_attendee(task.created_by, 'CHAIR')
            event['organizer'] = organizer
            already_included_users.append(task.created_by.pk)

        if task.responsible and task.responsible != task.created_by:
            attendee = self._create_attendee(task.responsible)
            event.add('attendee', attendee)
            already_included_users.append(task.responsible.pk)

        if task.followers.all().count() > 0:
            for user in task.followers.all():
                if user.pk in already_included_users:
                    continue

                attendee = self._create_attendee(user)
                event.add('attendee', attendee)

        if task.tags.all().count() > 0:
            for tag in task.tags.all():
                event.add('categories', tag)
        return event

    def _create_calendar(self, project, queryset):
        prodid = settings.CHAOS_CALENDAR_PRODID if hasattr(settings, 'CHAOS_CALENDAR_PRODID') else 'chaoscal'
        cal_version = settings.CHAOS_CALENDAR_VERSION if hasattr(settings, 'CHAOS_CALENDAR_VERSION') else '2.0'
        cal = Calendar()
        cal.add('prodid', prodid)
        cal.add('version', cal_version)
        cal['summary'] = project.name
        for task in queryset:
            event = self._create_event(task)
            cal.add_component(event)
        return cal

    def export(self, queryset):
        # TODO: we need to make sure this works on python 2 as well!!!
        if queryset.count() <= 0:
            return None
        project = queryset[0].project
        calendar = self._create_calendar(project, queryset)
        output_path = self.get_output_path()
        with codecs.open(output_path, 'wb') as output_file:
            output_file.write(calendar.to_ical())
        return output_path

    def respond(self, queryset):
        # TODO: we need to make sure this works on python 2 as well!!!
        try:
            logger.info('exporting tasks to ical')

            output_file = self.export(queryset)
            with open(output_file, 'rb') as out:
                output = out.read()
            download_name = self.get_download_name(output_file)
            response = HttpResponse(content_type=self.content_type, charset='UTF-8')
            response['Content-Disposition'] = 'attachment; filename=%s' % download_name
            response.write(output)

            logger.info('exporting tasks to ical finished.')
            return response
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while exporting tasks to ical. error: %s', message, exc_info=True)  # noqa
            return Responses.server_error()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-01-12 17:40
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('chaos', '0003_auto_20180104_1742'),
    ]

    operations = [
        migrations.CreateModel(
            name='Entrepreneur',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(help_text='Entrepreneur Name', max_length=200, verbose_name='Name')),
                ('nickname', models.CharField(max_length=200, null=True, verbose_name='Nickname')),
                ('document_number', models.CharField(max_length=64, null=True, verbose_name='Document Number')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Entrepreneur',
                'verbose_name_plural': 'Entrepreneurs',
            },
        ),
        migrations.CreateModel(
            name='EntrepreneurHasEnterprise',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('enterprise', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='chaos.Enterprise')),
                ('entrepreneur', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='chaos.Entrepreneur')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='entrepreneur',
            name='enterprises',
            field=models.ManyToManyField(related_name='entrepreneurs', through='chaos.EntrepreneurHasEnterprise', to='chaos.Enterprise'),
        ),
    ]



# coding: utf-8
from django.db.models.signals import pre_save
from django.dispatch import receiver
from .models import (
    Enterprise,
    Project,
    Task
)


@receiver(pre_save, sender=Enterprise)
@receiver(pre_save, sender=Project)
@receiver(pre_save, sender=Task)
def calculate_900913_geometry(sender, instance, *args, **kwargs):
    """
    Calculates and sets the 900913 geometry if
    regular geometry is given
    """

    if instance.geometry:
        instance.geometry_900913 = instance.geometry.transform(900913, clone=True)

    if hasattr(instance, 'extent') and hasattr(instance, 'extent_900913') and instance.extent:
        instance.extent_90013 = instance.extent.transform(900913, clone=True)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import chaos.models
import colorfield.fields
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields
import recurrence.fields
import taggit.managers
import uuid
from chaos.data_migrations import create_task_status


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('text', models.TextField(verbose_name='Text')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'ordering': ('-date_created',),
                'verbose_name': 'Comment',
                'verbose_name_plural': 'Comments',
            },
        ),
        migrations.CreateModel(
            name='Enterprise',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Enterprise',
                'verbose_name_plural': 'Enterprises',
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('name', models.CharField(max_length=64, verbose_name='Project')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('template', models.BooleanField(default=False, verbose_name='Is Template?')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('enterprise', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='projects', to='chaos.Enterprise', verbose_name='Enterprise')),
                ('responsible', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='projects', to=settings.AUTH_USER_MODEL)),
                ('tags', taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
            ],
            options={
                'verbose_name': 'Project',
                'verbose_name_plural': 'Projects',
            },
        ),
        migrations.CreateModel(
            name='ProjectType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('code', models.CharField(max_length=32, unique=True, verbose_name='Code')),
                ('description', models.CharField(max_length=128, null=True, verbose_name='Description')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Project Type',
                'verbose_name_plural': 'Project Types',
            },
        ),
        migrations.CreateModel(
            name='Reminder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('start_date_field', models.CharField(default=b'start_date', max_length=128, verbose_name='Start Date Field')),
                ('object_id', models.PositiveIntegerField()),
                ('time_delta', models.DurationField(verbose_name='Time Delta')),
                ('repeat', recurrence.fields.RecurrenceField(null=True, verbose_name='Repeat')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reminders', to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Reminder',
                'verbose_name_plural': 'Reminders',
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('start_date', models.DateTimeField(null=True, verbose_name='Start Date')),
                ('end_date', models.DateTimeField(null=True, verbose_name='End Date')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, verbose_name='UUID')),
                ('weight', models.IntegerField(default=1, verbose_name='Weight')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('description', models.TextField(null=True, verbose_name='Description')),
                ('lft', models.PositiveIntegerField(db_index=True, editable=False)),
                ('rght', models.PositiveIntegerField(db_index=True, editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(db_index=True, editable=False)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('parent', mptt.fields.TreeForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='chaos.Task')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to='chaos.Project')),
                ('responsible', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Task',
                'verbose_name_plural': 'Tasks',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='TaskAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='chaos.Task')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaskStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='Date Updated')),
                ('code', models.CharField(max_length=32, unique=True, verbose_name='Code')),
                ('description', models.CharField(max_length=128, null=True, verbose_name='Description')),
                ('color', colorfield.fields.ColorField(default=b'#A4A4A4', max_length=18)),
                ('is_system_status', models.BooleanField(default=False, verbose_name='Is System?')),
                ('is_closing_status', models.BooleanField(default=False, verbose_name='Is Closing Status?')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Task Status',
                'verbose_name_plural': 'Task Statuses',
            },
        ),
        migrations.AddField(
            model_name='task',
            name='status',
            field=models.ForeignKey(default=chaos.models.default_task_status, on_delete=django.db.models.deletion.CASCADE, to='chaos.TaskStatus', verbose_name='Status'),
        ),
        migrations.AddField(
            model_name='task',
            name='tags',
            field=taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='project',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='chaos.ProjectType', verbose_name='Type'),
        ),
        migrations.AddField(
            model_name='comment',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments', to='chaos.Task'),
        ),
        migrations.RunPython(create_task_status)
    ]



# coding: utf-8
from django.db.models import Q
import rest_framework_filters as filters
from rest_framework_filters.filterset import LOOKUP_SEP
from .models import (
    Entrepreneur,
    Enterprise,
    ProjectType,
    Project,
    TaskStatus,
    Task,
    Comment,
)
from datetime import datetime


class TagsFilter(filters.CharFilter):
    def filter(self, qs, value):
        if value:
            tags = [tag.strip() for tag in value.split(',')]
            qs = qs.filter(tags__name__in=tags).distinct()
        return qs


class EntrepreneurFilter(filters.FilterSet):

    class Meta:
        model = Entrepreneur
        fields = {
            'id': ['exact'],
            'name': ['exact', 'icontains'],
            'nickname': ['exact', 'icontains'],
            'document_number': ['exact', 'icontains'],
        }


class EnterpriseFilter(filters.FilterSet):

    entrepreneurs = filters.RelatedFilter(
        EntrepreneurFilter,
        field_name='entrepreneurs',
        queryset=Entrepreneur.objects.all()
    )

    class Meta:
        model = Enterprise
        fields = {
            'name': ['exact', 'startswith', 'icontains'],
            'description': ['icontains'],
        }


class ProjectTypeFilter(filters.FilterSet):

    class Meta:
        model = ProjectType
        fields = {
            'id': ['exact'],
            'code': ['exact', 'startswith', 'icontains'],
            'description': ['exact', 'startswith', 'icontains']
        }


class ProjectFilter(filters.FilterSet):

    class Meta:
        model = Project
        fields = {
            'id': ['exact'],
            'enterprise': ['exact'],
            'name': ['exact', 'startswith', 'icontains'],
            'type': ['exact'],
            'responsible': ['exact'],
            'description': ['icontains'],
        }


class TaskStatusFilter(filters.FilterSet):

    class Meta:
        model = TaskStatus
        fields = {
            'id': ['exact'],
            'code': ['exact', 'startswith', 'icontains'],
            'description': ['exact', 'startswith', 'icontains'],
            'is_system_status': ['exact'],
            'is_closing_status': ['exact']
        }


class TaskFilter(filters.FilterSet):

    start_date = filters.CharFilter(name='start_date', method='filter_start_date')
    end_date = filters.CharFilter(name='end_date', method='filter_end_date')
    root_task = filters.BooleanFilter(name='parent', method='filter_is_root_task')

    status = filters.RelatedFilter(
        TaskStatusFilter,
        name='status',
        queryset=TaskStatus.objects.all()
    )
    tags = TagsFilter(name="tags")

    def filter_start_date(self, qs, name, value):
        datetime_object = datetime.strptime(value, '%d/%m/%Y %H:%M')
        return qs.filter(Q(start_date__gte=datetime_object) | Q(end_date__gte=datetime_object))

    def filter_end_date(self, qs, name, value):
        datetime_object = datetime.strptime(value, '%d/%m/%Y %H:%M')
        return qs.filter(Q(end_date__lte=datetime_object) | Q(start_date__lte=datetime_object))

    def filter_is_root_task(self, qs, name, value):
        lookup = LOOKUP_SEP.join([name, 'isnull'])
        return qs.filter(**{lookup: value})

    class Meta:
        model = Task
        fields = {
            'id': ['exact'],
            'project': ['exact'],
            'root_task': ['exact'],
            'parent': ['exact'],
            'responsible': ['exact'],
            'status': ['exact'],
            'name': ['exact', 'startswith', 'icontains'],
            'description': ['icontains'],
        }


class CommentFilter(filters.FilterSet):

    class Meta:
        model = Comment
        fields = {
            'id': ['exact'],
            'created_by': ['exact'],
            'task': ['exact'],
            'text': ['icontains']
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-04-11 14:27
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chaos', '0006_auto_20180205_0749'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='extent',
            field=django.contrib.gis.db.models.fields.PolygonField(null=True, srid=4326, verbose_name='Extent'),
        ),
        migrations.AddField(
            model_name='task',
            name='extent_900913',
            field=django.contrib.gis.db.models.fields.PolygonField(null=True, srid=900913, verbose_name='Geometry 900913'),
        ),
        migrations.AlterField(
            model_name='task',
            name='geometry',
            field=django.contrib.gis.db.models.fields.PointField(null=True, srid=4326, verbose_name='Geometry'),
        ),
        migrations.AlterField(
            model_name='task',
            name='geometry_900913',
            field=django.contrib.gis.db.models.fields.PointField(null=True, srid=900913, verbose_name='Geometry 900913'),
        ),
    ]



# -*- coding: utf-8
import os
import logging
from django.apps import AppConfig
logger = logging.getLogger(__name__)


class ChaosConfig(AppConfig):
    name = 'chaos'

    def _create_directory(self, path):
        """
        Safely creates a directory for you.
        """
        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except Exception as ex:
                message = ex.message if hasattr(ex, 'message') else ''
                logger.error(
                    'error while creating directory %s. error: %s',
                    path,
                    message,
                    exc_info=True
                )

    def ready(self):
        from .receivers import *  # noqa
        from django.conf import settings
        CHAOS_EXPORT_PATH = (settings.CHAOS_EXPORT_PATH
                             if hasattr(settings, 'CHAOS_EXPORT_PATH')
                             else os.path.join(settings.MEDIA_ROOT, 'export', 'chaos'))
        self._create_directory(CHAOS_EXPORT_PATH)



# coding: utf-8
from django.dispatch import Signal

chaos_instance_updated = Signal(providing_args=['instance', 'user', 'data'])
chaos_instance_deleted = Signal(providing_args=['instance', 'user'])
task_status_changed = Signal(providing_args=['task', 'status', 'changes', 'user'])



# coding: utf-8
from collections import OrderedDict
from django.contrib.auth import get_user_model
from rest_framework import serializers
from rest_framework.reverse import reverse
from taggit_serializer.serializers import (
    TagListSerializerField,
    TaggitSerializer
)
from file_context.serializers import (
    FilesMixInSerializer,
)
from common.serializers import LinkSerializer
from .models import (
    Entrepreneur,
    Enterprise,
    ProjectType,
    Project,
    TaskStatus,
    Task,
    TaskAttachment,
    Reminder,
    Comment,
)
User = get_user_model()


class EntrepreneurAutoCompleteSerializer(serializers.ModelSerializer):

    """Auto complete serializer"""

    class Meta:
        model = Entrepreneur
        fields = (
            'id',
            'name'
        )


class EnterpriseSerializer(LinkSerializer,
                           FilesMixInSerializer):

    entrepreneurs = serializers.SerializerMethodField()
    project_count = serializers.SerializerMethodField()
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_entrepreneurs(self, obj):
        return [{'id': e.id, 'name': e.name} for e in obj.entrepreneurs.all()]

    def get_project_count(self, obj):
        return obj.projects.all().count()

    def get_links(self, obj):
        request = self.context.get('request')
        detailURL = reverse('project-detail', kwargs={'pk': obj.pk}, request=request)
        projectsURL = reverse('project-list', request=request) + '?enterprise={0}'.format(obj.pk)
        return OrderedDict([
            ('self', detailURL),
            ('projects', projectsURL),
        ])

    class Meta:

        model = Enterprise
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'name',
            'description',
            'entrepreneurs',
            'files',
            'links',
            'project_count',
            'geometry',
        )
        extra_kwargs = {
            'projects': {'read_only': True}
        }


class EntrepreneurSerializer(LinkSerializer,
                             FilesMixInSerializer):

    enterprises = EnterpriseSerializer(many=True, read_only=True)

    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_links(self, obj):
        request = self.context.get('request')
        detailURL = reverse('entrepreneur-detail', kwargs={'pk': obj.pk}, request=request)  # noqa
        return OrderedDict([
            ('self', detailURL),
        ])

    class Meta:
        model = Entrepreneur
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'name',
            'nickname',
            'document_number',
            'description',
            'enterprises',
            'files',
            'links'
        )


class ProjectTypeSerializer(LinkSerializer):
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    class Meta:

        model = ProjectType
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'code',
            'description',
        )


class ProjectSerializer(TaggitSerializer,
                        LinkSerializer,
                        FilesMixInSerializer):

    score = serializers.SerializerMethodField()
    full_name = serializers.SerializerMethodField()
    type_name = serializers.SerializerMethodField()
    responsible_name = serializers.SerializerMethodField()
    enterprise_name = serializers.SerializerMethodField()
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_score(self, obj):
        return obj.score

    def get_responsible_name(self, obj):
        if obj.responsible:
            return obj.responsible.get_username()
        else:
            return ''

    def get_type_name(self, obj):
        return obj.type.code

    def get_full_name(self, obj):
        return obj.full_name

    def get_enterprise_name(self, obj):

        if obj.enterprise:
            return obj.enterprise.name
        return ''

    def get_links(self, obj):

        request = self.context.get('request')
        detailURL = reverse('project-detail', kwargs={'pk': obj.pk}, request=request)
        tasksURL = reverse('task-list', request=request) + '?project={0}'.format(obj.pk)
        if obj.enterprise:
            enterpriseURL = reverse('enterprise-detail', kwargs={'pk': obj.enterprise.pk}, request=request)
        else:
            enterpriseURL = ''
        return OrderedDict([
            ('self', detailURL),
            ('enterprise', enterpriseURL),
            ('tasks', tasksURL),
            ('clone', detailURL + '/clone'),
            ('score', detailURL + '/score'),
        ])

    tags = TagListSerializerField()

    class Meta:

        model = Project
        fields = (
            'id',
            'created_by',
            'enterprise',
            'enterprise_name',
            'type',
            'type_name',
            'responsible',
            'responsible_name',
            'name',
            'full_name',
            'description',
            'template',
            'files',
            'tags',
            'links',
            'score',
            'geometry',
        )


class TaskStatusSerializer(LinkSerializer):
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_links(self, obj):

        request = self.context.get('request')
        detailUrl = reverse('task-status-detail', kwargs={'pk': obj.pk}, request=request)
        return {
            'self': detailUrl,
        }

    class Meta:
        model = TaskStatus
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'code',
            'description',
            'color',
            'is_system_status',
            'is_closing_status',
            'links',
        )


class ReminderSerializer(LinkSerializer):

    user_name = serializers.SerializerMethodField()
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_links(self, obj):
        return {}

    def get_user_name(self, obj):
        if obj.user:
            return obj.user.get_username()
        else:
            return ''

    class Meta:
        model = Reminder
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'content_type',
            'object_id',
            'user',
            'user_name',
            'is_recurring',
            'reminder_dates'
        )


class TaskSerializer(TaggitSerializer,
                     LinkSerializer,
                     FilesMixInSerializer):

    condition = serializers.SerializerMethodField()
    status_detail = serializers.SerializerMethodField()
    project_name = serializers.SerializerMethodField()
    responsible_name = serializers.SerializerMethodField()
    project_name = serializers.SerializerMethodField()
    parent_name = serializers.SerializerMethodField()

    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_project_name(self, obj):
        if obj.project:
            return obj.project.full_name
        else:
            return ''

    def get_parent_name(self, obj):
        if obj.parent:
            return obj.parent.name
        return ''

    def get_responsible_name(self, obj):
        if obj.responsible:
            return obj.responsible.get_username()
        return ''

    def get_status_detail(self, obj):
        status = obj.status
        if not status:
            return {}
        return {
            'code': status.code,
            'color': status.color,
            'is_system_status': status.is_system_status,
            'is_closing_status': status.is_closing_status
        }

    def get_condition(self, obj):
        return obj.condition

    def get_links(self, obj):

        request = self.context.get('request')
        detailUrl = reverse(
            'task-detail',
            kwargs={
                'pk': obj.pk
            },
            request=request
        )
        projectUrl = reverse(
            'project-detail',
            kwargs={
                'pk': obj.project_id
            },
            request=request
        )
        commentsUrl = reverse('comment-list', request=request) + '?task={0}'.format(obj.pk)
        return OrderedDict([
            ('self', detailUrl),
            ('project', projectUrl),
            ('comments', commentsUrl),
            ('change_status', detailUrl + '/change_status'),
            ('attach', detailUrl + '/attach'),
            ('detach', detailUrl + '/detach'),
            ('clone', detailUrl + '/clone'),
            ('score', detailUrl + '/score'),
            ('export', detailUrl + '/export'),
        ])

    tags = TagListSerializerField()

    class Meta:

        model = Task
        fields = (
            'id',
            'created_by',
            'date_created',
            'date_updated',
            'start_date',
            'end_date',
            'project',
            'project_name',
            'parent',
            'responsible',
            'responsible_name',
            'project_name',
            'parent_name',
            'status',
            'status_detail',
            'weight',
            'name',
            'description',
            'condition',
            'tags',
            'files',
            'links',
            'geometry'
        )


class CommentSerializer(LinkSerializer,
                        FilesMixInSerializer):

    author = serializers.SerializerMethodField()
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_author(self, obj):
        if obj.created_by:
            return obj.created_by.get_username()

    def get_links(self, obj):
        request = self.context.get('request')
        detailUrl = reverse(
            'comment-detail',
            kwargs={
                'pk': obj.pk
            },
            request=request
        )
        taskUrl = reverse(
            'task-detail',
            kwargs={
                'pk': obj.task_id
            },
            request=request
        )
        return OrderedDict([
            ('self', detailUrl),
            ('task', taskUrl),
        ])

    class Meta:
        model = Comment
        fields = (
            'id',
            'author',
            'created_by',
            'date_created',
            'date_updated',
            'task',
            'text',
            'files',
            'links'
        )


class TaskAttachmentSerializer(LinkSerializer):

    model_name = serializers.SerializerMethodField()
    created_by = serializers.PrimaryKeyRelatedField(
        read_only=True,
        default=serializers.CurrentUserDefault()
    )

    def get_model_name(self, obj):
        return '{0}.{1}'.format(obj.content_type.app_label, obj.content_type.name)

    class Meta:
        model = TaskAttachment
        fields = (
            'id',
            'task',
            'created_by',
            'content_type',
            'object_id',
            'model_name'
        )


class TaskChangeSerializer(serializers.Serializer):

    task = TaskSerializer()
    initial_target = TaskSerializer()
    action = serializers.CharField(max_length=64)
    status = TaskStatusSerializer()


class ListAllocationSerializer(serializers.Serializer):

    """
    Serializer that allows you to control
    filter for allocations
    """

    project = serializers.PrimaryKeyRelatedField(
        queryset=Project.objects.all(),
        required=False
    )

    min_date = serializers.DateField(
        required=False
    )

    max_date = serializers.DateField(
        required=False
    )


class AllocationSerializer(serializers.Serializer):

    user = serializers.SerializerMethodField()
    user_id = serializers.SerializerMethodField()
    date_point = serializers.DateTimeField()
    tasks = TaskSerializer(many=True, read_only=True)
    weight = serializers.SerializerMethodField()

    def get_user(self, obj):
        if obj.user:
            return obj.user.get_username()
        else:
            return ''

    def get_user_id(self, obj):
        if obj.user:
            return obj.user.pk
        else:
            return ''

    def get_weight(self, obj):
        return obj.weight


class AttachDetachSerializer(serializers.Serializer):
    """
    Serializer that validates and helps
    to construct the required request data
    for attach and detach endpoints
    """

    model_name = serializers.CharField()
    model_id = serializers.IntegerField()

    class Meta:
        fields = '__all__'


class ChangeStatusSerializer(serializers.Serializer):

    """
    Serializer that validates and helps
    to construct the required request data
    for change-status endpoint
    """

    status = serializers.IntegerField()
    dry_run = serializers.BooleanField(required=False)

    class Meta:
        fields = '__all__'


class ProjectCloneSerializer(serializers.Serializer):

    new_name = serializers.CharField()
    new_enterprise = serializers.PrimaryKeyRelatedField(
        queryset=Enterprise.objects.all(),
        required=False,
        allow_null=True
    )
    new_type = serializers.PrimaryKeyRelatedField(
        queryset=ProjectType.objects.all(),
        required=False,
        allow_null=True
    )
    new_responsible = serializers.PrimaryKeyRelatedField(
        queryset=User.objects.all(),
        required=False,
        allow_null=True
    )
    new_date = serializers.DateTimeField(
        required=False,
        allow_null=True
    )
    new_status = serializers.PrimaryKeyRelatedField(
        queryset=TaskStatus.objects.all(),
        required=False,
        allow_null=True
    )
    keep_tasks = serializers.BooleanField(default=True)
    keep_comments = serializers.BooleanField(default=False)
    keep_attachments = serializers.BooleanField(default=False)
    keep_tags = serializers.BooleanField(default=True)

    class Meta:
        fields = '__all__'


class TaskCloneSerializer(serializers.Serializer):

    new_project = serializers.PrimaryKeyRelatedField(
        queryset=Project.objects.filter(template=False),
        required=False,
        allow_null=True,
    )
    new_node = serializers.PrimaryKeyRelatedField(
        queryset=Task.objects.all(),
        required=False,
        allow_null=True,
    )
    new_status = serializers.PrimaryKeyRelatedField(
        queryset=TaskStatus.objects.all(),
        required=False,
        allow_null=True,
    )
    new_responsible = serializers.PrimaryKeyRelatedField(
        queryset=User.objects.all(),
        required=False,
        allow_null=True,
    )

    new_date = serializers.DateTimeField(
        required=False,
        allow_null=True
    )
    keep_comments = serializers.BooleanField(default=False)
    keep_attachments = serializers.BooleanField(default=False)
    keep_tags = serializers.BooleanField(default=True)

    class Meta:
        fields = '__all__'



# coding: utf-8
from rest_framework import routers
from file_context.viewsets import FileViewSet
from .viewsets_autocomplete import (
    EntrepreneurAutoCompleteViewSet,
)
from .viewsets import (
    EntrepreneurViewSet,
    EnterpriseViewSet,
    ProjectTypeViewSet,
    ProjectViewSet,
    ProjectTemplateViewSet,
    TaskStatusViewSet,
    TaskViewSet,
    CommentViewSet,
    AllocationViewSet,
    ReminderViewSet,
)


router = routers.DefaultRouter(trailing_slash=False)
router.register(r'entrepreneurs', EntrepreneurViewSet)
router.register(r'enterprises', EnterpriseViewSet)
router.register(r'project-types', ProjectTypeViewSet)
router.register(r'projects', ProjectViewSet)
router.register(r'templates', ProjectTemplateViewSet, base_name='templates')
router.register(r'task-status', TaskStatusViewSet, base_name='task-status')
router.register(r'tasks', TaskViewSet)
router.register(r'comments', CommentViewSet)
router.register(r'allocations', AllocationViewSet, base_name='allocations')
router.register(r'files', FileViewSet)
router.register(r'reminders', ReminderViewSet)
router.register(r'ac/entrepreneurs', EntrepreneurAutoCompleteViewSet)



# coding: utf-8
"""
This module responsibility is to hold
classes and functions that are related
with cloning projects and tasks.
"""
import logging
from .models import (
    Project,
    Task,
    TaskStatus,
    TaskAttachment,
    Comment,
    Reminder,
)
logger = logging.getLogger(__name__)
from django.db.models import Min
from datetime import timedelta

def clone_project(source,
                  new_name,
                  new_enterprise=None,
                  new_type=None,
                  new_responsible=None,
                  new_date=None,
                  new_status=None,
                  keep_tasks=False,
                  keep_comments=False,
                  keep_attachments=False,
                  keep_tags=False):
    try:
        if not source:
            raise ValueError('source is mandatory')
        if not new_name:
            raise ValueError('new_name is mandatory')
        logger.info('cloning project %s into %s', source, new_name)

        new_enterprise = new_enterprise if new_enterprise else source.enterprise
        new_type = new_type if new_type else source.type
        new_status = new_status if new_status else TaskStatus.objects.get(code='OPEN')
        new_responsible = new_responsible if new_responsible else source.responsible
        new = Project()
        new.enterprise = new_enterprise
        new.name = new_name
        new.description = source.description
        new.type = new_type
        new.responsible = new_responsible
        new.save()

        if keep_tags:
            logger.info('cloning project %s tags', source)
            for tag in source.tags.all():
                new.tags.add(tag)

        if keep_tasks:
            logger.info('cloning project %s tasks', source)
            for task in source.root_tasks.all():
                clone_task(
                    task,
                    new_project=new,
                    new_date=new_date,
                    new_status=new_status,
                    keep_comments=keep_comments,
                    keep_attachments=keep_attachments,
                    keep_tags=keep_tags
                )

        return new
    except Exception as ex:
        logger.error(
            'error while cloning project %s. error: %s',
            source,
            ex.message,
            exc_info=True
        )
        return None


def clone_task(source,
               new_project=None,
               new_node=None,
               new_status=None,
               new_date=None,
               new_responsible=None,
               keep_comments=False,
               keep_attachments=False,
               keep_tags=False):
    if not source:
        raise ValueError('source is mandatory')
    if not new_project and not new_node:
        raise ValueError('new_project and new_node cannot be null. you must choose one to be the target of the copied task.')  # noqa
    if new_node and new_project.pk != new_node.project.pk:
        raise ValueError('new_node must be contained in new_project')
    logger.info('cloning task %s', source)
    if new_project:
        project = new_project
    else:
        if new_node.project:
            project = new_node.project
        else:
            project = source.project

    project = new_project if new_project else source.project
    parent = new_node  # this does not coerce because new_node == None is also valid
    responsible = new_responsible if new_responsible else source.responsible

    status = new_status if new_status else source.status    

    start_date = new_date if new_date else source.start_date
    minimal_date_start = source.project.tasks.aggregate(Min('start_date'))
    minimal_date_end = source.project.tasks.aggregate(Min('end_date'))

    root_start_date = None
    root_end_date = None

    if start_date:
        if minimal_date_start['start_date__min']:
            if source.start_date:
                delta = source.start_date - minimal_date_start['start_date__min']
                absolute_delta = timedelta(seconds=delta.total_seconds() * -1 if delta.total_seconds() < 0 else delta.total_seconds())
                root_start_date = start_date + absolute_delta

    if root_start_date and source.end_date:
        delta = source.end_date - source.start_date
        root_end_date = root_start_date + delta
    elif minimal_date_end['end_date__min']:
        if source.end_date:
            delta = source.end_date - minimal_date_end['end_date__min']
            absolute_delta = timedelta(seconds=delta.total_seconds() * -1 if delta.total_seconds() < 0 else delta.total_seconds())
            root_end_date = start_date + absolute_delta

    new = Task(
        project=project,
        parent=parent,
        responsible=responsible,
        status=status,
        name=source.name,
        description=source.description,
        start_date=root_start_date,
        end_date=root_end_date
    )
    new.save()
    if keep_comments:
        logger.info('cloning task %s comments', source)
        for c in source.comments.all():
            new_comment = Comment(
                created_by=c.created_by,
                task=new,
                text=c.text
            )
            new_comment.save()
    if keep_attachments:
        logger.info('cloning task %s attachments', source)
        for a in source.attachments.all():
            new_attachment = TaskAttachment(
                created_by=a.created_by,
                task=new,
                content_object=a.content_object
            )
            new_attachment.save()
    if keep_tags:
        logger.info('cloning task %s tags', source)
        for tag in source.tags.all():
            new.tags.add(tag)

    logger.info('cloning task %s children', source)
    children = source.get_children()
    for child in children:
        child.clone(
            new_project=project,
            new_node=new,
            new_status=new_status,
            new_date=start_date,
            new_responsible=new_responsible,
            keep_comments=keep_comments,
            keep_attachments=keep_attachments
        )

    return new



# coding: utf-8
from rest_framework import viewsets
from .models import Entrepreneur
from .filters import EntrepreneurFilter
from .serializers import EntrepreneurAutoCompleteSerializer


class EntrepreneurAutoCompleteViewSet(viewsets.ReadOnlyModelViewSet):

    queryset = Entrepreneur.objects.all()
    serializer_class = EntrepreneurAutoCompleteSerializer
    filter_class = EntrepreneurFilter
    search_fields = (
        'id',
        'name'
    )



# coding: utf-8
import logging
from datetime import timedelta
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth import get_user_model
from django.db.models import ProtectedError
from django.utils.translation import ugettext_lazy as _
from rest_framework import viewsets
from rest_framework import status
from rest_framework.response import Response
from rest_framework.decorators import detail_route, list_route
from file_context.models import File
from .signals import (
    chaos_instance_updated,
    chaos_instance_deleted,
)
from .allocation import (
    AllocationHandler,
)
from .rest import (
    Responses
)
from .utils import (
    TaskHandler,
    get_model,
)
from .exporters import (
    XLSXTaskExport,
    ICALTaskExport,
)
from .models import (
    Entrepreneur,
    EntrepreneurHasEnterprise,
    Enterprise,
    ProjectType,
    Project,
    TaskStatus,
    Task,
    Comment,
    Reminder
)
from .filters import (
    EntrepreneurFilter,
    EnterpriseFilter,
    ProjectTypeFilter,
    ProjectFilter,
    TaskStatusFilter,
    TaskFilter,
    CommentFilter,
)
from .serializers import (
    EntrepreneurSerializer,
    EnterpriseSerializer,
    ProjectTypeSerializer,
    ProjectSerializer,
    TaskStatusSerializer,
    TaskSerializer,
    TaskAttachmentSerializer,
    TaskChangeSerializer,
    CommentSerializer,
    ReminderSerializer,
    AllocationSerializer,
    ChangeStatusSerializer,
    AttachDetachSerializer,
    ProjectCloneSerializer,
    TaskCloneSerializer,
    ListAllocationSerializer,
)
User = get_user_model()
logger = logging.getLogger(__name__)


def custom_update(self, viewset, request, pk=None, partial=False):
    """
    Custom update method. This does:
    * sends an update signal that includes the user - fired AFTER post_save.
    """
    try:
        user = request.user
        instance = self.get_object()
        if partial:
            result = super(viewset, self).update(request, **{'pk': pk, 'partial': partial})
        else:
            result = super(viewset, self).update(request, **{'pk': pk})
        chaos_instance_updated.send_robust(instance.__class__, instance=instance, user=user, data=request.data)
        return result
    except:
        logger.error('something bad happened while trying to update an object', exc_info=True)
        return Response('Cannot update object.')


def custom_destroy(self, viewset, request, pk=None):
    """
    Custom destroy
    * sends a signal that includes the user
    * prevents the destroy if the instance cannot be deleted
    """
    try:
        user = request.user
        instance = self.get_object()
        result = super(viewset, self).destroy(request, **{'pk': pk})
        chaos_instance_deleted.send_robust(instance.__class__, instance=instance, user=user)
        return result
    except ProtectedError:
        return Response('Cannot delete object because of protected foreign key relation',
                        status=status.HTTP_409_CONFLICT)


def handle_reminders(viewset, instance):
    request = viewset.request
    user = request.user
    reminders = request.data.get('reminders')
    task_ct = ContentType.objects.get_for_model(Task)
    if reminders:
        for r in reminders:
            try:
                delta = int(r['time_delta'])
            except:
                continue
            Reminder.objects.create(
                user=user,
                created_by=user,
                content_type=task_ct,
                object_id=instance.pk,
                time_delta=timedelta(days=delta)
            )


def handle_entrepreneurs(viewset, instance):
    request = viewset.request
    entrepreneurs = request.data.get('entrepreneurs')
    if not entrepreneurs:
        return
    instance.entrepreneurs.clear()
    for e_id in entrepreneurs:
        if not e_id:
            continue
        try:
            entrepreneur = Entrepreneur.objects.get(pk=e_id)
            EntrepreneurHasEnterprise.objects.create(
                entrepreneur=entrepreneur,
                enterprise=instance
            )
        except Entrepreneur.DoesNotExist:
            pass


def handle_files(viewset, instance):
    request = viewset.request
    files = request.data.get('files')
    if files:
        for file_id in files:
            try:
                file_instance = File.objects.get(pk=file_id)
                file_instance.attach(instance)
            except File.DoesNotExist:
                pass


class EntrepreneurViewSet(viewsets.ModelViewSet):

    queryset = Entrepreneur.objects.all()\
        .prefetch_related('enterprises')
    serializer_class = EntrepreneurSerializer
    filter_class = EntrepreneurFilter
    search_fields = (
        'name',
        'nickname',
        'document',
        'description',
    )

    def perform_create(self, serializer):
        super(EntrepreneurViewSet, self).perform_create(serializer)
        handle_files(self, serializer.instance)

    def partial_update(self, request, pk=None):
        return custom_update(self, EntrepreneurViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, EntrepreneurViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, EntrepreneurViewSet, request, pk)


class EnterpriseViewSet(viewsets.ModelViewSet):

    queryset = Enterprise.objects.all()\
        .prefetch_related('entrepreneurs')\
        .select_related('created_by')
    serializer_class = EnterpriseSerializer
    filter_class = EnterpriseFilter
    search_fields = (
        'id',
        'name',
        'description',
    )

    def perform_update(self, serializer):
        super(EnterpriseViewSet, self).perform_update(serializer)
        handle_entrepreneurs(self, serializer.instance)

    def perform_create(self, serializer):
        """
        overrides perform_create to automagically
        treat files in the request :D
        """
        super(EnterpriseViewSet, self).perform_create(serializer)
        handle_files(self, serializer.instance)
        handle_entrepreneurs(self, serializer.instance)

    def partial_update(self, request, pk=None):
        return custom_update(self, EnterpriseViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, EnterpriseViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, EnterpriseViewSet, request, pk)


class ProjectTypeViewSet(viewsets.ModelViewSet):

    queryset = ProjectType.objects.all()
    serializer_class = ProjectTypeSerializer
    filter_class = ProjectTypeFilter
    search_fields = (
        'id',
        'code',
        'description',
    )

    def partial_update(self, request, pk=None):
        return custom_update(self, ProjectTypeViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, ProjectTypeViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, ProjectTypeViewSet, request, pk)


class BaseProjectViewSet(object):

    queryset = Project.objects.all()\
        .select_related('created_by')\
        .select_related('type')\
        .select_related('responsible')\
        .select_related('enterprise')\
        .prefetch_related('tags')
    serializer_class = ProjectSerializer
    filter_class = ProjectFilter
    search_fields = (
        'id',
        'name',
        'type',
        'responsible',
        'description',
        'tags',
    )

    @detail_route(methods=['post'])
    def clone(self, request, pk=None):
        project = self.get_object()
        pcs = ProjectCloneSerializer(data=request.data)
        if not pcs.is_valid():
            return Response(pcs.errors, status=status.HTTP_400_BAD_REQUEST)
        try:
            new_project = project.clone(
                pcs.validated_data.get('new_name'),
                pcs.validated_data.get('new_enterprise'),
                pcs.validated_data.get('new_type'),
                pcs.validated_data.get('new_responsible'),
                pcs.validated_data.get('new_date'),
                pcs.validated_data.get('new_status'),
                pcs.validated_data.get('keep_tasks'),
                pcs.validated_data.get('keep_comments'),
                pcs.validated_data.get('keep_attachments'),
                pcs.validated_data.get('keep_tags')
            )
            if new_project:
                project_serializer = ProjectSerializer(new_project, context={'request': request})
                return Response(project_serializer.data, status=status.HTTP_201_CREATED)
            else:
                logger.error('an internal exception ocurred while cloning a project')
                return Responses.server_error(_('Something bad happened while cloning project.'))
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while cloning project %s. error: %s', project, message, exc_info=True)
            return Responses.server_error(message)

    @detail_route(methods=['get'])
    def score(self, request, pk=None):
        project = self.get_object()
        return Response({'project': project.pk, 'score': project.score})

    def perform_create(self, serializer):
        """
        overrides perform_create to automagically
        treat files in the request :D
        """
        super(BaseProjectViewSet, self).perform_create(serializer)
        handle_files(self, serializer.instance)

    def partial_update(self, request, pk=None):
        return custom_update(self, BaseProjectViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, BaseProjectViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, BaseProjectViewSet, request, pk)


class ProjectTemplateViewSet(BaseProjectViewSet, viewsets.ModelViewSet):

    queryset = Project.objects.filter(template=True)


class ProjectViewSet(BaseProjectViewSet, viewsets.ModelViewSet):

    queryset = Project.objects.filter(template=False)


class TaskStatusViewSet(viewsets.ModelViewSet):

    queryset = TaskStatus.objects.all()
    serializer_class = TaskStatusSerializer
    filter_class = TaskStatusFilter
    search_fields = (
        'id',
        'code',
        'description',
        'is_system_status',
        'is_closing_status'
    )

    def partial_update(self, request, pk=None):
        return custom_update(self, TaskStatusViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, TaskStatusViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, TaskStatusViewSet, request, pk)


class TaskViewSet(viewsets.ModelViewSet):

    queryset = Task.objects.all()\
        .select_related('created_by')\
        .select_related('status')\
        .select_related('project__enterprise')\
        .select_related('project')\
        .select_related('responsible')\
        .prefetch_related('comments')\
        .prefetch_related('tags')
    serializer_class = TaskSerializer
    filter_class = TaskFilter
    search_fields = (
        'id',
        'project',
        'parent',
        'responsible',
        'status',
        'name',
        'description',
        'tags',
    )
    exporters = {
        'XLSX': XLSXTaskExport,
        'ICAL': ICALTaskExport,
    }

    @list_route(methods=['get'])
    def export(self, request, *args, **kwargs):
        extension = request.query_params.get('extension', 'XLSX')
        extension = extension.upper()
        queryset = self.filter_queryset(self.get_queryset())
        if extension not in self.exporters:
            return Responses.client_error()
        try:
            exporter = self.exporters[extension](publish_url=True, request=request)
            return exporter.respond(queryset)
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while trying to export tasks to %s format. error: ', extension, message, exc_info=True)
            return Responses.server_error()

    @detail_route(methods=['get'])
    def score(self, request, pk=None):
        task = self.get_object()
        return Response({'task': task.pk, 'score': task.score})

    @detail_route(methods=['post'])
    def clone(self, request, pk=None):
        task = self.get_object()
        tcs = TaskCloneSerializer(data=request.data)
        if not tcs.is_valid():
            return Response(tcs.errors, status=status.HTTP_400_BAD_REQUEST)

        new_project = tcs.validated_data.get('new_project')
        new_node = tcs.validated_data.get('new_node')
        new_status = tcs.validated_data.get('new_status')
        new_date = tcs.validated_data.get('new_date')
        new_responsible = tcs.validated_data.get('new_responsible')
        keep_comments = tcs.validated_data.get('keep_comments')
        keep_attachments = tcs.validated_data.get('keep_attachments')
        keep_tags = tcs.validated_data.get('keep_tags')

        try:
            new_task = task.clone(
                new_project,
                new_node,
                new_status,
                new_date,
                new_responsible,
                keep_comments,
                keep_attachments,
                keep_tags
            )
            if new_task:
                task_serializer = TaskSerializer(new_task)
                return Response(
                    task_serializer.data,
                    status=status.HTTP_201_CREATED
                )
            else:
                logger.error('an internal exception ocurred while cloning a task')
                return Responses.server_error(_('Something bad happened while cloning a task.'))
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('failed to clone task %s. error: %s', task, message, exc_info=True)
            return Responses.server_error(message)

    @detail_route(methods=['post'])
    def attach(self, request, pk=None):
        task = self.get_object()
        ads = AttachDetachSerializer(data=request.data)
        if not ads.is_valid():
            return Response(ads.errors, status=status.HTTP_400_BAD_REQUEST)
        model_name = ads.validated_data.get('model_name')
        model_id = ads.validated_data.get('model_id')
        model_class = get_model(model_name)
        if not model_class:
            return Responses.not_found('model not found')
        try:
            user = request.user if not request.user.is_anonymous else None
            model = model_class.objects.get(pk=model_id)
            attachment = task.attach(model, user=user)
            att_serializer = TaskAttachmentSerializer(attachment)
            return Response(att_serializer.data)
        except model_class.DoesNotExist:
            logger.warn('%s model with id %s does not exist', model_name, model_id)
            return Responses.not_found()
        except Exception as ex:
            message = ex.message if hasattr(ex.message) else ''
            logger.error('error while attaching %s - %s to %s. error: %s', model_name, model_id, task, message)
            return Responses.server_error(message)

    @detail_route(methods=['post'])
    def detach(self, request, pk=None):
        task = self.get_object()
        ads = AttachDetachSerializer(data=request.data)
        if not ads.is_valid():
            return Response(ads.errors, status=status.HTTP_400_BAD_REQUEST)
        model_name = ads.validated_data.get('model_name')
        model_id = ads.validated_data.get('model_id')
        model_class = get_model(model_name)
        if not model_class:
            return Responses.not_found('model not found')
        try:
            model = model_class.objects.get(pk=model_id)
            task.detach(model)
            return Responses.success('Detach succeeded')
        except model_class.DoesNotExist:
            logger.warn('model %s with id %s does not exist', model_name, model_id)  # noqa
            return Responses.not_found()
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while detaching %s - %s from %s. error: %s', model_name, model_id, task, message)  # noqa
            return Responses.server_error(message)

    @detail_route(methods=['post'])
    def change_status(self, request, pk=None):
        task = self.get_object()
        css = ChangeStatusSerializer(data=request.data)
        if not css.is_valid():
            logger.info('invalid request for change_status')
            return Response(css.errors, status=status.HTTP_400_BAD_REQUEST)

        change = css.validated_data
        user = None if request.user.is_anonymous() else request.user
        try:
            new_status = TaskStatus.objects.get(pk=change.get('status'))
            handler = TaskHandler()
            changes = handler.to_status(
                task=task,
                status=new_status,
                dry_run=change.get('dry_run', True),
                user=user
            )
            if not changes:
                logger.info('change_status is not allowed. from %s to %s.', task.status, new_status)
                return Responses.succeeded('No changes detected, no changes executed.')

            logger.info('executing status changes')
            flat_changes = TaskHandler.flatten(changes)
            change_serializer = TaskChangeSerializer(flat_changes, many=True)
            return Response(data=change_serializer.data)

        except TaskStatus.DoesNotExist:
            logger.warn('status selected by the user does not exist.')
            return Responses.not_found()
        except Exception as ex:
            error_msg = getattr(ex, 'message', '')
            logger.error('something wrong happened while trying to change task status. %s', error_msg, exc_info=True)
            return Responses.server_error(error_msg)

    def perform_create(self, serializer):
        """
        overrides perform_create to automagically
        treat files in the request :D
        """
        super(TaskViewSet, self).perform_create(serializer)
        handle_files(self, serializer.instance)
        handle_reminders(self, serializer.instance)

    def partial_update(self, request, pk=None):
        return custom_update(self, TaskViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, TaskViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, TaskViewSet, request, pk)


class CommentViewSet(viewsets.ModelViewSet):

    queryset = Comment.objects.all()\
        .select_related('task')\
        .select_related('created_by')
    serializer_class = CommentSerializer
    filter_class = CommentFilter
    search_fields = (
        'id',
        'created_by',
        'task',
        'text',
    )

    def perform_create(self, serializer):
        """
        overrides perform_create to automagically
        treat files in the request :D
        """
        super(CommentViewSet, self).perform_create(serializer)
        handle_files(self, serializer.instance)

    def partial_update(self, request, pk=None):
        return custom_update(self, CommentViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, CommentViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, CommentViewSet, request, pk)


class ReminderViewSet(viewsets.ModelViewSet):

    queryset = Reminder.objects.all()
    serializer_class = ReminderSerializer
    search_fields = (
        'user',
        'content_type',
        'object_id',
        'is_recurring'
    )

    def partial_update(self, request, pk=None):
        return custom_update(self, ReminderViewSet, request, pk, partial=True)

    def update(self, request, pk=None):
        return custom_update(self, ReminderViewSet, request, pk)

    def destroy(self, request, pk=None):
        return custom_destroy(self, ReminderViewSet, request, pk)


class AllocationViewSet(viewsets.ViewSet):

    """
    Viewset that allows you to see how your resources
    are allocated
    """

    def list(self, request):
        try:
            # TODO: does not work
            list_allocation_serializer = ListAllocationSerializer(data=request.data)
            if not list_allocation_serializer.is_valid():
                return Response(data=list_allocation_serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            project = list_allocation_serializer.validated_data.get('project')
            min_date = list_allocation_serializer.validated_data.get('min_date')
            max_date = list_allocation_serializer.validated_data.get('max_date')

            users = User.objects.all()

            allocation_handler = AllocationHandler()
            allocations = allocation_handler.get_allocations(users, min_date, max_date, project)
            allocation_serial = AllocationSerializer(allocations, many=True)
            return Response(allocation_serial.data)
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while fetching list of allocations. Error: %s', message, exc_info=True)
            return Responses.server_error(message)

    def retrieve(self, request, pk=None):

        try:
            user = User.objects.get(pk=pk)
        except User.DoesNotExist:
            return Responses.not_found()

        try:
            las = ListAllocationSerializer(data=request.data)
            if not las.is_valid():
                return Response(data=las.errors, status=status.HTTP_400_BAD_REQUEST)

            project = las.validated_data.get('project')
            min_date = las.validated_data.get('min_date')
            max_date = las.validated_data.get('max_date')

            allocation_handler = AllocationHandler()
            allocations = allocation_handler.get_user_allocations(user, min_date, max_date, project)
            allocation_serializer = AllocationSerializer(allocations, many=True)
            return Response(allocation_serializer.data)
        except Exception as ex:
            message = ex.message if hasattr(ex, 'message') else ''
            logger.error('error while fetching allocation for user %s. Error: %s', user, message, exc_info=True)
            return Responses.server_error(message)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-01-19 18:57
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chaos', '0004_auto_20180112_1140'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='comment',
            options={'ordering': ('id',), 'verbose_name': 'Comment', 'verbose_name_plural': 'Comments'},
        ),
    ]



# codign: utf-8
import os
import uuid
import logging
from django.utils.translation import ugettext_lazy as _
from file_context.models import File
from .models import TaskStatus
from .signals import task_status_changed
logger = logging.getLogger(__name__)


def get_output_path(directory, filename):

    if not directory:
        raise ValueError('directory is mandatory')

    if not filename:
        raise ValueError('filename is mandatory')

    try:
        name, extension = filename.split('.')
    except:
        raise ValueError('filename must be composed of filename.extension')

    uid = uuid.uuid4().hex
    full_name = '{0}-{1}.{2}'.format(name, uid, extension)

    return os.path.join(directory, full_name)


def get_model(model_name):

    """
    model_name must be app.model
    """

    try:
        from django.db.models import get_model
    except:
        from django.apps import apps
        get_model = apps.get_model

    app, model = model_name.split('.')
    if not app or not model:
        raise ValueError('model_name is must be in the format app.model')
    return get_model(app, model)


def get_responsibles(task_changes, include_project=True):
    """
    Returns a unique list of users
    responsible for the tasks in these
    changes.
    """
    if not task_changes:
        return None
    changes = TaskHandler.flatten(task_changes)
    users = list([c.task.responsible for c in changes if c.task.responsible is not None])
    if include_project and task_changes.task.project.responsible:
        users.append(task_changes.task.project.responsible)
    return list(set(users))


class TaskChange(object):

    """
    this holds a list of task changes that
    needs to be done when talking about status.
    we gather them, so they can be applied later
    and trigger all kinds of side effects, like
    notifications.
    Task is the task that will be modified.
    action is a string that contains a description of
    the action itself, like opening, reopening or closing
    status is the new status of the task. since
    we can have custom open/closed status, this means
    we have to pass this status along.
    initial_target is the initial task that was operated
    on. since that target can cascade different changes
    up/down the tree, we maintain it, so we can give
    accurate notifications.
    """
    task = None
    action = None
    status = None
    initial_target = None
    related_changes = None

    def __init__(self, task, action, status, initial_target=None):
        self.task = task
        self.action = action
        self.status = status
        self.initial_target = initial_target
        self.related_changes = list()

    def __str__(self):
        return '%s task %s. related changes: %s' % (self.action, self.task, len(self.related_changes))

    def perform(self):
        logger.info(
            '%s: changing task %s status from %s to %s',
            self.action,
            self.task,
            self.task.status,
            self.status
        )
        self.task.status = self.status
        self.task.save()
        for related in self.related_changes:
            related.perform()


class TaskChangeDetector(object):

    def __init__(self,
                 allow_all=True):
        # allow_all determines if it's possible to
        # change states that have the same semantics,
        # like  from a (open) > b (open)
        self.allow_all = allow_all
        self.open_status = TaskStatus.objects.get(code='OPEN')
        self.closed_status = TaskStatus.objects.get(code='CLOSED')

    def detect(self, task, status):
        """
        performs a state change in the task.
        this will receive a new status and it
        will analyze it, and call the correct
        function for it.
        """
        if not task:
            logger.warn('detect cannot be run without task')
            raise ValueError('task is mandatory')
        if not status:
            logger.warn('detect cannot be run without status')
            raise ValueError('status is mandatory')

        if task.status.pk == status.pk:
            return None
        if status.is_closing_status:
            action = self._close
        else:
            action = self._open

        return action(task, status, task)

    def _open(self, task, status=None, initial_target=None):

        if not status:
            status = self.open_status

        if not task.status.is_closing_status:
            action = 'state change'
        else:
            action = 'opening'
        if not self.allow_all and not task.status.is_closing_status:
            # if this is already open, we just change the state
            # of this task
            logger.info(
                'task %s is already open on state %s, skipping change to %s',
                task,
                task.status,
                status
            )
            return None

        changes = TaskChange(task, action, status, initial_target)
        if task.parent and task.parent.status.is_closing_status:
            changes.related_changes.append(self._open(task.parent, status, initial_target))
        return changes

    def _close(self, task, status=None, initial_target=None):
        if not status:
            status = self.closed_status

        if task.status.is_closing_status:
            action = 'state change'
        else:
            action = 'closing'

        if not self.allow_all and task.status.is_closing_status:
            # if it's closed, we are not closing it again
            logger.info(
                'task %s is already closed on state %s, skipping change to %s',
                task,
                task.status,
                status
            )
            return None

        siblings_open = task.get_siblings().filter(status__is_closing_status=False)
        if siblings_open.count() == 0 and task.parent:
            # close parent and let him close everything
            # else
            logger.info(
                'all siblings from %s are closed. closing parent task %s',
                task,
                task.parent
            )
            return self._close(task.parent, status, initial_target)

        changes = TaskChange(task, action, status, initial_target)
        logger.info(
            'detected task change for %s task. changing from %s to %s state',
            task,
            task.status,
            status
        )
        for child_task in task.get_descendants().filter(status__is_closing_status=False):
            changes.related_changes.append(
                TaskChange(child_task, action, status, initial_target)
            )
        return changes


class TaskHandler(object):

    """
    This class allows you to operate
    certain actions on the tasks.
    it will discover the changes that
    are needed and will make them happen
    """

    def __init__(self, detector_class=None):
        if not detector_class:
            detector_class = TaskChangeDetector
        self.detector = detector_class()

    @classmethod
    def flatten(cls, change):
        """
        This method will flatten changes and return a list of TaskChange
        """
        changes = list()
        if not change:
            return changes
        changes.append(change)
        for related in change.related_changes:
            changes.extend(cls.flatten(related))
        return changes

    def _flatten(self, change):
        return self.__class__.flatten(change)

    def to_status(self, task, status, dry_run=False, user=None):
        """
        Changes the task to a certain status, closing or opening it
        depending on the status.
        """
        if not user:
            user = _('system')
        changes = self.detector.detect(task, status)
        if not dry_run:
            logger.info('TaskHandler not in dry run mode. Executing changes.')
            if changes:
                changes.perform()
                task_status_changed.send_robust(
                    sender=self.__class__,
                    task=task,
                    status=status,
                    changes=changes,
                    user=user
                )
            logger.info('TaskHandler done with change execution.')
        return changes



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-01-04 23:42
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('chaos', '0002_auto_20171127_0627'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='responsible',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='projects', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='task',
            name='responsible',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to=settings.AUTH_USER_MODEL),
        ),
    ]



# codign: utf-8
import logging
logger = logging.getLogger(__name__)


def _calculate_node_score(node):
    scores = list()
    if not node:
        logger.info('empty node, returning 0')
        return (0, 0, )

    children = node.get_children()
    if len(children) <= 0:
        total = node.weight
        if node.status.is_closing_status:
            done = node.weight
        else:
            done = 0
        return (done, total, )
    else:
        for child in children:
            scores.append(_calculate_node_score(child))

        closed = sum([s[0] for s in scores])
        total = sum([s[1] for s in scores])
        return (closed, total, )


def calculate_node_score(node):
    score = _calculate_node_score(node)
    total = score[1]
    if total == 0:
        total = 1
    return float(score[0]) / float(total)


def calculate_project_score(project):

    """
    Calculates a projects score, based
    on how many open/closed tasks it has
    """

    if not project:
        raise ValueError('project is required to calculate score')
    root_tasks = project.root_tasks.all()
    if root_tasks.count() <= 0:
        return 0.0

    scores = [calculate_node_score(r) for r in root_tasks]
    return sum(scores) / float(len(scores))



# coding: utf-8
"""
This module is responsible to figure out the allocation
for each user in hours/days.
"""
from datetime import timedelta
from django.db.models import Min, Max
from django.contrib.auth import get_user_model
from .models import Task
User = get_user_model()


class Allocation(object):

    user = None
    date_point = None
    tasks = None

    def __init__(self, user, date_point, tasks):
        """
        Initializes a new allocation record.
        """
        self.user = user
        self.date_point = date_point
        self.tasks = tasks

    @property
    def weight(self):
        return sum([task.weight for task in self.tasks])


class AllocationHandler(object):

    def get_allocations(self, users_queryset=None, min_date=None, max_date=None, project=None, include_without_responsible=True):
        allocations = list()
        if not users_queryset:
            users_queryset = User.objects.all()

        for user in users_queryset:
            allocations.extend(self.get_user_allocations(user, min_date, max_date, project))
        if include_without_responsible:
            allocations.extend(self.get_user_allocations(None, min_date, max_date, project))
        return allocations

    def get_user_allocations(self, user, min_date=None, max_date=None, project=None):
        temp = dict()
        allocs = list()
        tasks = Task.objects.filter(
            responsible=user,
            status__is_closing_status=False,
        ).exclude(
            start_date=None,
            end_date=None
        )
        if project:
            tasks = tasks.filter(project=project)
        if not min_date:
            # the usage of .all here is because
            # when you call all, it will clone the current
            # queryset, allowing you to add more queries to it,
            # without changing what is already filtered.
            min_date = tasks.all().aggregate(Min('start_date'))
            min_date = min_date.get('start_date__min')
        if not max_date:
            max_date = tasks.all().aggregate(Max('end_date'))
            max_date = max_date.get('end_date__max')
        if not min_date and not max_date:
            return allocs
        # inclusive range here, because range = 0 means
        # the start date and the last date we want is
        # max_date, so we add 1o max_date
        tasks.filter(start_date__gte=min_date, end_date__lte=max_date)
        for t in tasks:
            for day in range(0, t.duration.days + 1):
                current_datetime = t.start_date + timedelta(days=day)
                if temp.get(current_datetime):
                    temp[current_datetime].append(t)
                else:
                    temp[current_datetime] = [t]
        return [Allocation(user=user, date_point=dt, tasks=ts) for dt, ts in temp.items()]
